@@ -51,6 +51,7 @@
 #include <errno.h>
 #include <string.h>
 #include <sys/stat.h>
+#include <sys/mman.h>
 
 #include "madmodule.h"
 #include "pymadfile.h"
@@ -61,10 +62,11 @@
 #endif
 
 #define ERROR_MSG_SIZE  512
-#define MAD_BUF_SIZE    4096 /* should be a multiple of 4 >= 4096 */
+#define MAD_BUF_SIZE    (5*8192) /* should be a multiple of 4 >= 4096 and large enough to capture possible 
+                                    junk at beginning of MP3 file*/
 
 /* local helpers */
-static unsigned long calc_total_time(PyObject *);
+static unsigned long    calc_total_time(PyObject *);
 static signed short int madfixed_to_short(mad_fixed_t);
 
 PyTypeObject py_madfile_t = {
@@ -128,7 +130,7 @@ PyObject * py_madfile_new(PyObject * self, PyObject * args) {
     /* bufsiz must be an integer multiple of 4 */
     if ((n = bufsiz % 4)) bufsiz -= n;
     if (bufsiz <= 4096) bufsiz = 4096;
-  
+
     mf = PyObject_NEW(py_madfile, &py_madfile_t);
     Py_INCREF(fobject);
     mf->fobject = fobject;
@@ -141,7 +143,7 @@ PyObject * py_madfile_new(PyObject * self, PyObject * args) {
     mad_timer_reset(&mf->timer);
   
     mf->framecount = 0;
-  
+
     mf->buffy = malloc(bufsiz * sizeof(unsigned char));
     mf->bufsiz = bufsiz;
 
@@ -179,8 +181,8 @@ static void py_madfile_dealloc(PyObject * self, PyObject * args) {
     PyObject_DEL(self);
 }
 
-/* FIXME: some MP3s have bogus headers, and the only way to calculate length
- * is to decode the whole thing */
+/* Calculates length of MP3 by stepping through the frames and summing up
+ * the duration of each frame. */
 static unsigned long
 calc_total_time(PyObject *self) {
     mad_timer_t timer;
@@ -207,10 +209,40 @@ calc_total_time(PyObject *self) {
         fnum = PyInt_AsLong(o);
         Py_DECREF(o);
 	r = fstat(fnum, &buf);
-	if (r == 0 && MAD_FRAME(self).header.bitrate)
-	    r = buf.st_size * 8 / MAD_FRAME(self).header.bitrate * 1000;
-        else
-	    r = -1;
+
+	/* Figure out actual length of file by stepping through it.
+	 * This is a stripped down version of how madplay does it. */
+	void *ptr = mmap(0, buf.st_size, PROT_READ, MAP_SHARED, fnum, 0);
+	if(!ptr) {
+            fprintf(stderr, "mmap failed, can't calculate length");
+            return -1;
+	}
+
+        mad_timer_t time = mad_timer_zero;
+        struct mad_stream stream;
+        struct mad_header header;
+
+        mad_stream_init(&stream);
+        mad_header_init(&header);
+
+        mad_stream_buffer(&stream, ptr, buf.st_size);
+
+        while (1) {
+            if(mad_header_decode(&header, &stream) == -1) {
+                if( MAD_RECOVERABLE(stream.error))
+                    continue;
+                else
+                    break;
+            }
+
+            mad_timer_add(&time, header.duration);
+        }
+
+        if (munmap(ptr, buf.st_size) == -1) {
+            return -1;
+        }
+
+        r = time.seconds * 1000;
     }
     return r;
 }
@@ -255,10 +287,11 @@ madfixed_to_short(mad_fixed_t sample) {
 
 static PyObject * 
 py_madfile_read(PyObject * self, PyObject * args) {
-    PyObject * pybuf, * tuple; /* output objects */
+    PyObject * pybuf;    /* return object containing output buffer*/
     char * buffy; /* output buffer */
     unsigned int i, size;
     int nextframe = 0;
+    int result;
     char errmsg[ERROR_MSG_SIZE];
 
     /* if we are at EOF, then return None */
@@ -328,7 +361,6 @@ py_madfile_read(PyObject * self, PyObject * args) {
                 Py_INCREF(Py_None);
                 return Py_None;
             }
-            // FIXME: deallocate o_read?
             memcpy(readstart, o_buffer, readsize);
             Py_DECREF(o_read);
 
@@ -367,14 +399,17 @@ py_madfile_read(PyObject * self, PyObject * args) {
 	 * this case one can call again mad_frame_decode() in order to
 	 * skip the faulty part and resync to the next frame.
 	 */
-	if (mad_frame_decode(&MAD_FRAME(self), &MAD_STREAM(self))) {
+        Py_BEGIN_ALLOW_THREADS
+	result = mad_frame_decode(&MAD_FRAME(self), &MAD_STREAM(self));
+        Py_END_ALLOW_THREADS
+        if (result) {
 	    if (MAD_RECOVERABLE(MAD_STREAM(self).error)) {
 		/* FIXME: prefer to return an error string to the caller
-		 * rather than print to stderr
+		 * rather than print to stderr 
 		fprintf(stderr, "mad: recoverable frame level error: %s\n",
 		        mad_stream_errorstr(&MAD_STREAM(self)));
 		fflush(stderr);
-		*/
+                */
 		/* go onto the next frame */
 		nextframe = 1;
 	    } else {
@@ -393,6 +428,8 @@ py_madfile_read(PyObject * self, PyObject * args) {
 	
     } while (nextframe);
   
+    Py_BEGIN_ALLOW_THREADS
+    
     /* Accounting.  The computed frame duration is in the frame header
      * structure.  It is expressed as a fixed point number whose data
      * type is mad_timer_t.  It is different from the fixed point
@@ -407,24 +444,16 @@ py_madfile_read(PyObject * self, PyObject * args) {
     /* Once decoded, the frame can be synthesised to PCM samples.
      * No errors are reported by mad_synth_frame() */
     mad_synth_frame(&MAD_SYNTH(self), &MAD_FRAME(self));
+
+    Py_END_ALLOW_THREADS
   
     /* Create the buffer to store the PCM samples in, so python can
      * use it.  We do 4 pointer increments per sample in the buffer,
      * so make it 4 times as big as the number of samples */
     size = MAD_SYNTH(self).pcm.length * 4;
     pybuf = PyBuffer_New(size);
-  
-    /* create a tuple object so we can get at the stuff in buf */
-    tuple = PyTuple_New(1);
-    Py_INCREF(pybuf);
-    PyTuple_SET_ITEM(tuple, 0, pybuf);
-    if (!PyArg_ParseTuple(tuple, "t#", &buffy, &size)) {
-	PyErr_SetString(PyExc_TypeError, "borken buffer tuple!");
-	return NULL;
-    }
-    /* no longer use tuple */
-    Py_DECREF(tuple);
-  
+    PyObject_AsWriteBuffer(pybuf, &buffy, &size);
+
     /* die if we don't have the space */
     if (size < MAD_SYNTH(self).pcm.length * 4) {
 	PyErr_SetString(PyExc_MemoryError, "allocated buffer too small");
@@ -442,16 +471,26 @@ py_madfile_read(PyObject * self, PyObject * args) {
 
 	/* left channel */
 	sample = madfixed_to_short(MAD_SYNTH(self).pcm.samples[0][i]);
+#ifdef BIGENDIAN
+	*(buffy++) = sample >> 8;
+	*(buffy++) = sample & 0xFF;
+#else
 	*(buffy++) = sample & 0xFF;
 	*(buffy++) = sample >> 8;
+#endif
 
 	/* right channel. 
 	 * if the decoded stream is monophonic then the right channel
 	 * is the same as the left one */
 	if (MAD_NCHANNELS(&MAD_FRAME(self).header) == 2)
 	    sample = madfixed_to_short(MAD_SYNTH(self).pcm.samples[1][i]);
+#ifdef BIGENDIAN
+	*(buffy++) = sample >> 8;
+	*(buffy++) = sample & 0xFF;
+#else
 	*(buffy++) = sample & 0xFF;
 	*(buffy++) = sample >> 8;
+#endif
     }
   
     Py_END_ALLOW_THREADS
@@ -543,7 +582,6 @@ py_madfile_seek_time(PyObject * self, PyObject * args) {
 
     mad_timer_set(&MAD_TIMER(self), 0, pos, 1000);
 
-    Py_INCREF(Py_None);
     return Py_None;
 }
 
